"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
class Pipeline {
    producer;
    components = Array();
    constructor(producer) {
        this.producer = producer;
    }
    static from(producer) {
        return new Pipeline(producer);
    }
    then(component) {
        this.components.push(component);
        return this;
    }
    async finally(final) {
        const state = this.components.map(() => ({
            buffer: Array(),
            costUsage: 0,
            driving: 0,
        }));
        const isFull = (idx) => state[idx].costUsage >= this.components[idx].costBuffer;
        const push = (idx, val) => {
            state[idx].buffer.push(val);
            state[idx].costUsage += this.components[idx].inputCost(val);
            ensureDrivingComponent(idx);
        };
        const ensureDrivingComponent = async (idx) => {
            const isFirst = idx === 0;
            const isLast = idx === state.length - 1;
            if (state[idx].driving >= this.components[idx].concurrency) {
                return;
            }
            state[idx].driving++;
            while ((isLast || !isFull(idx + 1)) && state[idx].buffer.length) {
                const input = state[idx].buffer.shift();
                state[idx].costUsage -= this.components[idx].inputCost(input);
                // If shifting has now made this component's buffer not full, drive the previous component.
                // NOTE: Shifting may not always make it non-full.
                if (!isFull(idx)) {
                    if (isFirst) {
                        ensureDrivingProducer();
                    }
                    else {
                        ensureDrivingComponent(idx - 1);
                    }
                }
                const output = await this.components[idx].handler(input);
                if (output !== undefined) {
                    if (isLast) {
                        final(output);
                    }
                    else {
                        push(idx + 1, output);
                    }
                }
            }
            state[idx].driving--;
        };
        let drivingProducer = false;
        const ensureDrivingProducer = async () => {
            if (drivingProducer) {
                return;
            }
            drivingProducer = true;
            while (!isFull(0)) {
                const output = await this.producer();
                push(0, output);
            }
            drivingProducer = false;
        };
        ensureDrivingProducer();
    }
}
exports.default = Pipeline;
